@@ -17,14 +17,21 @@
 #include <sys/ioctl.h>  // IWYU pragma: keep
 #endif
 
+#include <algorithm>
+#include <functional>
+#include <iterator>
 #include <memory>
+#include <mutex>
 #include <sstream>
 #include <string>
+#include <tuple>
 #include <type_traits>
+#include <unordered_map>
 #include <vector>
 
 #include "fallback.h"  // IWYU pragma: keep
-#include "signal.h"    // IWYU pragma: keep
+#include "maybe.h"
+#include "signal.h"  // IWYU pragma: keep
 
 // Define a symbol we can use elsewhere in our code to determine if we're being built on MS Windows
 // under Cygwin.
@@ -101,7 +108,22 @@ static_assert(false, "Neither NAME_MAX nor MAXNAMELEN is defined!");
 #endif
 #endif
 
-enum escape_string_style_t { STRING_STYLE_SCRIPT, STRING_STYLE_URL, STRING_STYLE_VAR };
+// PATH_MAX may not exist.
+#ifndef PATH_MAX
+#ifdef MAXPATHLEN
+#define PATH_MAX MAXPATHLEN
+#else
+/// Fallback length of MAXPATHLEN. Hopefully a sane value.
+#define PATH_MAX 4096
+#endif
+#endif
+
+enum escape_string_style_t {
+    STRING_STYLE_SCRIPT,
+    STRING_STYLE_URL,
+    STRING_STYLE_VAR,
+    STRING_STYLE_REGEX,
+};
 
 // Flags for unescape_string functions.
 enum {
@@ -158,9 +180,20 @@ enum selection_direction_t {
 ///
 /// will print the string 'fish: Pi = 3.141', given that debug_level is 1 or higher, and that
 /// program_name is 'fish'.
-void __attribute__((noinline)) debug(int level, const char *msg, ...)
+void __attribute__((noinline)) debug_impl(int level, const char *msg, ...)
     __attribute__((format(printf, 2, 3)));
-void __attribute__((noinline)) debug(int level, const wchar_t *msg, ...);
+void __attribute__((noinline)) debug_impl(int level, const wchar_t *msg, ...);
+
+/// The verbosity level of fish. If a call to debug has a severity level higher than \c debug_level,
+/// it will not be printed.
+extern int debug_level;
+
+inline bool should_debug(int level) { return level <= debug_level; }
+
+#define debug(level, ...)                                            \
+    do {                                                             \
+        if (should_debug((level))) debug_impl((level), __VA_ARGS__); \
+    } while (0)
 
 /// Exits without invoking destructors (via _exit), useful for code after fork.
 [[noreturn]] void exit_without_destructors(int code);
@@ -171,6 +204,9 @@ extern struct termios shell_modes;
 /// The character to use where the text has been truncated. Is an ellipsis on unicode system and a $
 /// on other systems.
 extern wchar_t ellipsis_char;
+/// The character or string to use where text has been truncated (ellipsis if possible, otherwise
+/// ...)
+extern const wchar_t *ellipsis_str;
 
 /// Character representing an omitted newline at the end of text.
 extern wchar_t omitted_newline_char;
@@ -178,10 +214,6 @@ extern wchar_t omitted_newline_char;
 /// Character used for the silent mode of the read command
 extern wchar_t obfuscation_read_char;
 
-/// The verbosity level of fish. If a call to debug has a severity level higher than \c debug_level,
-/// it will not be printed.
-extern int debug_level;
-
 /// How many stack frames to show when a debug() call is made.
 extern int debug_stack_frames;
 
@@ -191,9 +223,16 @@ extern bool g_profiling_active;
 /// Name of the current program. Should be set at startup. Used by the debug function.
 extern const wchar_t *program_name;
 
-/// Set to false at run-time if it's been determined we can't trust the last modified timestamp on
-/// the tty.
-extern bool has_working_tty_timestamps;
+/// Set to false if it's been determined we can't trust the last modified timestamp on the tty.
+extern const bool has_working_tty_timestamps;
+
+/// A list of all whitespace characters
+extern const wcstring whitespace;
+extern const char *whitespace_narrow;
+
+bool is_whitespace(const wchar_t input);
+bool is_whitespace(const wcstring &input);
+inline bool is_whitespace(const wchar_t *input) { return is_whitespace(wcstring(input)); }
 
 /// This macro is used to check that an argument is true. It is a bit like a non-fatal form of
 /// assert. Instead of exiting on failure, the current function is ended at once. The second
@@ -211,12 +250,12 @@ extern bool has_working_tty_timestamps;
 // from within a `switch` block. As of the time I'm writing this oclint doesn't recognize the
 // `__attribute__((noreturn))` on the exit_without_destructors() function.
 // TODO: we use C++11 [[noreturn]] now, does that change things?
-#define FATAL_EXIT()                       \
-    {                                      \
-        char exit_read_buff;               \
-        show_stackframe(L'E');             \
+#define FATAL_EXIT()                                \
+    {                                               \
+        char exit_read_buff;                        \
+        show_stackframe(L'E');                      \
         ignore_result(read(0, &exit_read_buff, 1)); \
-        exit_without_destructors(1);       \
+        exit_without_destructors(1);                \
     }
 
 /// Exit the program at once after emitting an error message and stack trace if possible.
@@ -263,8 +302,18 @@ extern bool has_working_tty_timestamps;
 /// See https://developer.gnome.org/glib/stable/glib-I18N.html#N-:CAPS
 #define N_(wstr) wstr
 
-/// Test if a list of stirngs contains a particular string.
-bool contains(const wcstring_list_t &list, const wcstring &str);
+/// Test if a collection contains a value.
+template <typename Col, typename T2>
+bool contains(const Col &col, const T2 &val) {
+    return std::find(std::begin(col), std::end(col), val) != std::end(col);
+}
+
+/// Append a vector \p donator to the vector \p receiver.
+template <typename T>
+void vec_append(std::vector<T> &receiver, std::vector<T> &&donator) {
+    receiver.insert(receiver.end(), std::make_move_iterator(donator.begin()),
+                    std::make_move_iterator(donator.end()));
+}
 
 /// Print a stack trace to stderr.
 void show_stackframe(const wchar_t msg_level, int frame_count = 100, int skip_levels = 0);
@@ -284,6 +333,7 @@ int fgetws2(wcstring *s, FILE *f);
 wcstring str2wcstring(const char *in);
 wcstring str2wcstring(const char *in, size_t len);
 wcstring str2wcstring(const std::string &in);
+wcstring str2wcstring(const std::string &in, size_t len);
 
 /// Returns a newly allocated multibyte character string equivalent of the specified wide character
 /// string.
@@ -297,15 +347,118 @@ std::string wcs2string(const wcstring &input);
 /// Test if a string prefixes another. Returns true if a is a prefix of b.
 bool string_prefixes_string(const wcstring &proposed_prefix, const wcstring &value);
 bool string_prefixes_string(const wchar_t *proposed_prefix, const wcstring &value);
+bool string_prefixes_string(const wchar_t *proposed_prefix, const wchar_t *value);
+bool string_prefixes_string(const char *proposed_prefix, const std::string &value);
+bool string_prefixes_string(const char *proposed_prefix, const char *value);
 
 /// Test if a string is a suffix of another.
 bool string_suffixes_string(const wcstring &proposed_suffix, const wcstring &value);
 bool string_suffixes_string(const wchar_t *proposed_suffix, const wcstring &value);
+bool string_suffixes_string_case_insensitive(const wcstring &proposed_suffix,
+                                             const wcstring &value);
 
 /// Test if a string prefixes another without regard to case. Returns true if a is a prefix of b.
 bool string_prefixes_string_case_insensitive(const wcstring &proposed_prefix,
                                              const wcstring &value);
 
+/// Case-insensitive string search, templated for use with both std::string and std::wstring.
+/// Modeled after std::string::find().
+/// \param fuzzy indicates this is being used for fuzzy matching and case insensitivity is
+/// expanded to include symbolic characters (#3584).
+/// \return the offset of the first case-insensitive matching instance of `needle` within
+/// `haystack`, or `string::npos()` if no results were found.
+template <typename T>
+size_t ifind(const T &haystack, const T &needle, bool fuzzy = false) {
+    using char_t = typename T::value_type;
+    auto locale = std::locale();
+
+    std::function<bool(char_t, char_t)> icase_eq;
+
+    if (!fuzzy) {
+        icase_eq = [&locale](char_t c1, char_t c2) {
+            return std::toupper(c1, locale) == std::toupper(c2, locale);
+        };
+    } else {
+        icase_eq = [&locale](char_t c1, char_t c2) {
+            // This `ifind()` call is being used for fuzzy string matching. Further extend case
+            // insensitivity to treat `-` and `_` as equal (#3584).
+
+            // The two lines below were tested to be 27% faster than
+            //      (c1 == '_' || c1 == '-') && (c2 == '-' || c2 == '_')
+            // while returning no false positives for all (c1, c2) combinations in the printable
+            // range (0x20-0x7E). It might return false positives outside that range, but fuzzy
+            // comparisons are typically called for file names only, which are unlikely to have
+            // such characters and this entire function is 100% broken on unicode so there's no
+            // point in worrying about anything outside of the ANSII range.
+            // ((c1 == Literal<char_t>('_') || c1 == Literal<char_t>('-')) &&
+            // ((c1 ^ c2) == (Literal<char_t>('-') ^ Literal<char_t>('_'))));
+
+            // One of the following would be an illegal comparison between a char and a wchar_t.
+            // However, placing them behind a constexpr gate results in the elision of the if
+            // statement and the incorrect branch, with the compiler's SFINAE support suppressing
+            // any errors in the branch not taken.
+            if (sizeof(char_t) == sizeof(char)) {
+                return std::toupper(c1, locale) == std::toupper(c2, locale) ||
+                ((c1 == '_' || c1 == '-') &&
+                ((c1 ^ c2) == ('-' ^ '_')));
+            } else {
+                return std::toupper(c1, locale) == std::toupper(c2, locale) ||
+                ((c1 == L'_' || c1 == L'-') &&
+                ((c1 ^ c2) == (L'-' ^ L'_')));
+            }
+        };
+    }
+
+    auto result =
+        std::search(haystack.begin(), haystack.end(), needle.begin(), needle.end(), icase_eq);
+    if (result != haystack.end()) {
+        return result - haystack.begin();
+    }
+    return T::npos;
+}
+
+/// Split a string by a separator character.
+wcstring_list_t split_string(const wcstring &val, wchar_t sep);
+
+/// Join a list of strings by a separator character.
+wcstring join_strings(const wcstring_list_t &vals, wchar_t sep);
+
+/// Support for iterating over a newline-separated string.
+template <typename Collection>
+class line_iterator_t {
+    // Storage for each line.
+    Collection storage;
+
+    // The collection we're iterating. Note we hold this by reference.
+    const Collection &coll;
+
+    // The current location in the iteration.
+    typename Collection::const_iterator current;
+
+public:
+    /// Construct from a collection (presumably std::string or std::wcstring).
+    line_iterator_t(const Collection &coll) : coll(coll), current(coll.cbegin()) {}
+
+    /// Access the storage in which the last line was stored.
+    const Collection &line() const {
+        return storage;
+    }
+
+    /// Advances to the next line. \return true on success, false if we have exhausted the string.
+    bool next() {
+        if (current == coll.end())
+            return false;
+        auto newline_or_end = std::find(current, coll.cend(), '\n');
+        storage.assign(current, newline_or_end);
+        current = newline_or_end;
+
+        // Skip the newline.
+        if (current != coll.cend())
+            ++current;
+        return true;
+    }
+};
+
 enum fuzzy_match_type_t {
     // We match the string exactly: FOOBAR matches FOOBAR.
     fuzzy_match_exact = 0,
@@ -322,6 +475,9 @@ enum fuzzy_match_type_t {
     // We match a substring of the string: OOBA matches FOOBAR.
     fuzzy_match_substring,
 
+    // We match a substring of the string: ooBA matches FOOBAR.
+    fuzzy_match_substring_case_insensitive,
+
     // A subsequence match with insertions only: FBR matches FOOBAR.
     fuzzy_match_subsequence_insertions_only,
 
@@ -387,10 +543,39 @@ void assert_is_background_thread(const char *who);
 #define ASSERT_IS_BACKGROUND_THREAD_TRAMPOLINE(x) assert_is_background_thread(x)
 #define ASSERT_IS_BACKGROUND_THREAD() ASSERT_IS_BACKGROUND_THREAD_TRAMPOLINE(__FUNCTION__)
 
+// fish_mutex is a wrapper around std::mutex that tracks whether it is locked, allowing for checking
+// if the mutex is locked. It owns a boolean guarded by the lock that records whether the lock is
+// currently locked; this is only used by assertions for correctness.
+class fish_mutex_t {
+    std::mutex lock_{};
+    bool is_locked_{false};
+
+   public:
+    constexpr fish_mutex_t() = default;
+    ~fish_mutex_t() = default;
+
+    void lock() {
+        lock_.lock();
+        is_locked_ = true;
+    }
+
+    void unlock() {
+        is_locked_ = false;
+        lock_.unlock();
+    }
+
+    // assert that this lock (identified as 'who') is locked in the function 'caller'.
+    void assert_is_locked(const char *who, const char *caller) const;
+
+    // return the underlying std::mutex. Note the fish_mutex_t cannot track locks to the underlying
+    // mutex; do not use assert_is_locked() with this.
+    std::mutex &get_mutex() { return lock_; }
+};
+
 /// Useful macro for asserting that a lock is locked. This doesn't check whether this thread locked
 /// it, which it would be nice if it did, but here it is anyways.
-void assert_is_locked(void *mutex, const char *who, const char *caller);
-#define ASSERT_IS_LOCKED(x) assert_is_locked((void *)(&x), #x, __FUNCTION__)
+void assert_is_locked(const fish_mutex_t &m, const char *who, const char *caller);
+#define ASSERT_IS_LOCKED(x) (x).assert_is_locked(#x, __FUNCTION__)
 
 /// Format the specified size (in bytes, kilobytes, etc.) into the specified stringbuffer.
 wcstring format_size(long long sz);
@@ -465,13 +650,13 @@ char **make_null_terminated_array(const std::vector<std::string> &lst);
 // Helper class for managing a null-terminated array of null-terminated strings (of some char type).
 template <typename CharType_t>
 class null_terminated_array_t {
-    CharType_t **array;
+    CharType_t **array{NULL};
 
     // No assignment or copying.
     void operator=(null_terminated_array_t rhs);
     null_terminated_array_t(const null_terminated_array_t &);
 
-    typedef std::vector<std::basic_string<CharType_t> > string_list_t;
+    typedef std::vector<std::basic_string<CharType_t>> string_list_t;
 
     size_t size() const {
         size_t len = 0;
@@ -489,18 +674,30 @@ class null_terminated_array_t {
     }
 
    public:
-    null_terminated_array_t() : array(NULL) {}
+    null_terminated_array_t() = default;
+
     explicit null_terminated_array_t(const string_list_t &argv)
         : array(make_null_terminated_array(argv)) {}
 
     ~null_terminated_array_t() { this->free(); }
 
+    null_terminated_array_t(null_terminated_array_t &&rhs) : array(rhs.array) {
+        rhs.array = nullptr;
+    }
+
+    null_terminated_array_t operator=(null_terminated_array_t &&rhs) {
+        free();
+        array = rhs.array;
+        rhs.array = nullptr;
+    }
+
     void set(const string_list_t &argv) {
         this->free();
         this->array = make_null_terminated_array(argv);
     }
 
     const CharType_t *const *get() const { return array; }
+    CharType_t **get() { return array; }
 
     void clear() { this->free(); }
 };
@@ -509,70 +706,8 @@ class null_terminated_array_t {
 // null_terminated_array_t<char_t>.
 void convert_wide_array_to_narrow(const null_terminated_array_t<wchar_t> &arr,
                                   null_terminated_array_t<char> *output);
-
-class mutex_lock_t {
-   public:
-    pthread_mutex_t mutex;
-    mutex_lock_t() { DIE_ON_FAILURE(pthread_mutex_init(&mutex, NULL)); }
-
-    ~mutex_lock_t() { DIE_ON_FAILURE(pthread_mutex_destroy(&mutex)); }
-};
-
-// Basic scoped lock class.
-class scoped_lock {
-    pthread_mutex_t *lock_obj;
-    bool locked;
-
-    // No copying.
-    scoped_lock &operator=(const scoped_lock &) = delete;
-    scoped_lock(const scoped_lock &) = delete;
-
-   public:
-    scoped_lock(scoped_lock &&rhs) : lock_obj(rhs.lock_obj), locked(rhs.locked) {
-        // we acquire any locked state
-        // ensure the rhs doesn't try to unlock
-        rhs.locked = false;
-    }
-
-    void lock(void);
-    void unlock(void);
-    explicit scoped_lock(pthread_mutex_t &mutex);
-    explicit scoped_lock(mutex_lock_t &lock);
-    ~scoped_lock();
-};
-
-class rwlock_t {
-   public:
-    pthread_rwlock_t rwlock;
-    rwlock_t() { DIE_ON_FAILURE(pthread_rwlock_init(&rwlock, NULL)); }
-
-    ~rwlock_t() { DIE_ON_FAILURE(pthread_rwlock_destroy(&rwlock)); }
-
-    rwlock_t &operator=(const rwlock_t &) = delete;
-    rwlock_t(const rwlock_t &) = delete;
-};
-
-// Scoped lock class for rwlocks.
-class scoped_rwlock {
-    pthread_rwlock_t *rwlock_obj;
-    bool locked;
-    bool locked_shared;
-
-    // No copying.
-    scoped_rwlock &operator=(const scoped_lock &) = delete;
-    explicit scoped_rwlock(const scoped_lock &) = delete;
-
-   public:
-    void lock(void);
-    void unlock(void);
-    void lock_shared(void);
-    void unlock_shared(void);
-    // Upgrade shared lock to exclusive. Equivalent to `lock.unlock_shared(); lock.lock();`.
-    void upgrade(void);
-    explicit scoped_rwlock(pthread_rwlock_t &rwlock, bool shared = false);
-    explicit scoped_rwlock(rwlock_t &rwlock, bool shared = false);
-    ~scoped_rwlock();
-};
+typedef std::lock_guard<fish_mutex_t> scoped_lock;
+typedef std::lock_guard<std::recursive_mutex> scoped_rlock;
 
 // An object wrapping a scoped lock and a value
 // This is returned from owning_lock.acquire()
@@ -586,20 +721,25 @@ class scoped_rwlock {
 //
 template <typename DATA>
 class acquired_lock {
-    scoped_lock lock;
-    acquired_lock(mutex_lock_t &lk, DATA *v) : lock(lk), value(*v) {}
+    std::unique_lock<fish_mutex_t> lock;
+    acquired_lock(fish_mutex_t &lk, DATA *v) : lock(lk), value(v) {}
 
     template <typename T>
     friend class owning_lock;
 
+    DATA *value;
+
    public:
-    // No copying, move only
+    // No copying, move construction only
     acquired_lock &operator=(const acquired_lock &) = delete;
     acquired_lock(const acquired_lock &) = delete;
     acquired_lock(acquired_lock &&) = default;
     acquired_lock &operator=(acquired_lock &&) = default;
 
-    DATA &value;
+    DATA *operator->() { return value; }
+    const DATA *operator->() const { return value; }
+    DATA &operator*() { return *value; }
+    const DATA &operator*() const { return *value; }
 };
 
 // A lock that owns a piece of data
@@ -612,7 +752,7 @@ class owning_lock {
     owning_lock(owning_lock &&) = default;
     owning_lock &operator=(owning_lock &&) = default;
 
-    mutex_lock_t lock;
+    fish_mutex_t lock;
     DATA data;
 
    public:
@@ -650,6 +790,44 @@ class scoped_push {
     }
 };
 
+/// A helper class for managing and automatically closing a file descriptor.
+class autoclose_fd_t {
+    int fd_;
+
+   public:
+    // Closes the fd if not already closed.
+    void close();
+
+    // Returns the fd.
+    int fd() const { return fd_; }
+
+    // Returns the fd, transferring ownership to the caller.
+    int acquire() {
+        int temp = fd_;
+        fd_ = -1;
+        return temp;
+    }
+
+    // Resets to a new fd, taking ownership.
+    void reset(int fd) {
+        if (fd == fd_) return;
+        close();
+        fd_ = fd;
+    }
+
+    autoclose_fd_t(const autoclose_fd_t &) = delete;
+    void operator=(const autoclose_fd_t &) = delete;
+    autoclose_fd_t(autoclose_fd_t &&rhs) : fd_(rhs.fd_) { rhs.fd_ = -1; }
+
+    void operator=(autoclose_fd_t &&rhs) {
+        close();
+        std::swap(this->fd_, rhs.fd_);
+    }
+
+    explicit autoclose_fd_t(int fd = -1) : fd_(fd) {}
+    ~autoclose_fd_t() { close(); }
+};
+
 /// Appends a path component, with a / if necessary.
 void append_path_component(wcstring &path, const wcstring &component);
 
@@ -658,7 +836,7 @@ wcstring vformat_string(const wchar_t *format, va_list va_orig);
 void append_format(wcstring &str, const wchar_t *format, ...);
 void append_formatv(wcstring &str, const wchar_t *format, va_list ap);
 
-#ifdef __cpp_lib_make_unique
+#ifdef HAVE_STD__MAKE_UNIQUE
 using std::make_unique;
 #else
 /// make_unique implementation
@@ -706,14 +884,19 @@ wcstring escape_string(const wchar_t *in, escape_flags_t flags,
 wcstring escape_string(const wcstring &in, escape_flags_t flags,
                        escape_string_style_t style = STRING_STYLE_SCRIPT);
 
+/// \return a string representation suitable for debugging (not for presenting to the user). This
+/// replaces non-ASCII characters with either tokens like <BRACE_SEP> or <\xfdd7>. No other escapes
+/// are made (i.e. this is a lossy escape).
+wcstring debug_escape(const wcstring &in);
+
 /// Expand backslashed escapes and substitute them with their unescaped counterparts. Also
 /// optionally change the wildcards, the tilde character and a few more into constants which are
 /// defined in a private use area of Unicode. This assumes wchar_t is a unicode character set.
 
 /// Given a null terminated string starting with a backslash, read the escape as if it is unquoted,
-/// appending to result. Return the number of characters consumed, or 0 on error.
-size_t read_unquoted_escape(const wchar_t *input, wcstring *result, bool allow_incomplete,
-                            bool unescape_special);
+/// appending to result. Return the number of characters consumed, or none() on error.
+maybe_t<size_t> read_unquoted_escape(const wchar_t *input, wcstring *result, bool allow_incomplete,
+                                     bool unescape_special);
 
 /// Unescapes a string in-place. A true result indicates the string was unescaped, a false result
 /// indicates the string was unmodified.
@@ -780,6 +963,19 @@ void assert_is_not_forked_child(const char *who);
 #define ASSERT_IS_NOT_FORKED_CHILD_TRAMPOLINE(x) assert_is_not_forked_child(x)
 #define ASSERT_IS_NOT_FORKED_CHILD() ASSERT_IS_NOT_FORKED_CHILD_TRAMPOLINE(__FUNCTION__)
 
+/// Detect if we are Windows Subsystem for Linux by inspecting /proc/sys/kernel/osrelease
+/// and checking if "Microsoft" is in the first line.
+/// See https://github.com/Microsoft/WSL/issues/423 and Microsoft/WSL#2997
+constexpr bool is_windows_subsystem_for_linux() {
+    // This function is called after fork() and before exec() in postfork.cpp. Make sure we
+    // don't allocate any memory here!
+#ifdef WSL
+    return true;
+#else
+    return false;
+#endif
+}
+
 extern "C" {
 __attribute__((noinline)) void debug_thread_error(void);
 }
@@ -843,6 +1039,22 @@ static const wchar_t *enum_to_str(T enum_val, const enum_map<T> map[]) {
     return NULL;
 };
 
+template <typename... Args>
+using tuple_list = std::vector<std::tuple<Args...>>;
+
+// Given a container mapping one X to many Y, return a list of {X,Y}
+template <typename X, typename Y>
+inline tuple_list<X, Y> flatten(const std::unordered_map<X, std::vector<Y>> &list) {
+    tuple_list<X, Y> results(list.size() * 1.5);  // just a guess as to the initial size
+    for (auto &kv : list) {
+        for (auto &v : kv.second) {
+            results.emplace_back(std::make_tuple(kv.first, v));
+        }
+    }
+
+    return results;
+}
+
 void redirect_tty_output();
 
 // Minimum allowed terminal size and default size if the detected size is not reasonable.
@@ -866,9 +1078,13 @@ enum {
     STATUS_CMD_OK = 0,
     /// The status code used for failure exit in a command (but not if the args were invalid).
     STATUS_CMD_ERROR = 1,
+    /// The status code used for invalid arguments given to a command. This is distinct from valid
+    /// arguments that might result in a command failure. An invalid args condition is something
+    /// like an unrecognized flag, missing or too many arguments, an invalid integer, etc. But
+    STATUS_INVALID_ARGS = 2,
+
     /// The status code used when a command was not found.
     STATUS_CMD_UNKNOWN = 127,
-
     /// TODO: Figure out why we have two distinct failure codes for when an external command cannot
     /// be run.
     ///
@@ -883,10 +1099,6 @@ enum {
     STATUS_ILLEGAL_CMD = 123,
     /// The status code used when `read` is asked to consume too much data.
     STATUS_READ_TOO_MUCH = 122,
-    /// The status code used for invalid arguments given to a command. This is distinct from valid
-    /// arguments that might result in a command failure. An invalid args condition is something
-    /// like an unrecognized flag, missing or too many arguments, an invalid integer, etc. But
-    STATUS_INVALID_ARGS = 121,
 };
 
 /* Normally casting an expression to void discards its value, but GCC
@@ -895,10 +1107,43 @@ enum {
    and __extension__ to work around the problem, if the workaround is
    known to be needed.  */
 #if 3 < __GNUC__ + (4 <= __GNUC_MINOR__)
-# define ignore_result(x) \
-    (__extension__ ({ __typeof__ (x) __x = (x); (void) __x; }))
+#define ignore_result(x)         \
+    (__extension__({             \
+        __typeof__(x) __x = (x); \
+        (void)__x;               \
+    }))
 #else
-# define ignore_result(x) ((void) (x))
+#define ignore_result(x) ((void)(x))
 #endif
 
+// Custom hash function used by unordered_map/unordered_set when key is const
+#ifndef CONST_WCSTRING_HASH
+#define CONST_WCSTRING_HASH 1
+namespace std {
+template <>
+struct hash<const wcstring> {
+    std::size_t operator()(const wcstring &w) const {
+        std::hash<wcstring> hasher;
+        return hasher((wcstring)w);
+    }
+};
+}  // namespace std
+#endif
+
+/// Get the absolute path to the fish executable itself
+std::string get_executable_path(const char *fallback);
+
+/// A RAII wrapper for resources that don't recur, so we don't have to create a separate RAII
+/// wrapper for each function. Avoids needing to call "return cleanup()" or similar / everywhere.
+struct cleanup_t {
+private:
+    const std::function<void()> cleanup;
+public:
+    cleanup_t(std::function<void()> exit_actions)
+        : cleanup{exit_actions} {}
+    ~cleanup_t() {
+        cleanup();
+    }
+};
+
 #endif
